use crate::error::Error;
use std::{
    convert::TryFrom,
    fmt::{Display, Formatter},
};

#[derive(Clone)]
pub struct SectionInfo<I> {
    pub file_size: u64,
    pub data_size: u64,
    pub data_index: Option<I>, // some form of indexing to the source data (Relocatable's data)
    pub offset: u64,
}

#[derive(Clone)]
pub enum Section<I> {
    Text(SectionInfo<I>),
    Data(SectionInfo<I>, bool), // readonly bool
    Bss(SectionInfo<I>),
}

impl<I> Section<I> {
    pub fn file_size(&self) -> Result<u64, Error> {
        match self {
            Section::Text(s) => Ok(s.file_size),
            Section::Data(s, _) => Ok(s.file_size),
            Section::Bss(s) => Ok(s.file_size),
        }
    }

    pub fn data_size(&self) -> Result<u64, Error> {
        match self {
            Section::Text(s) => Ok(s.data_size),
            Section::Data(s, _) => Ok(s.data_size),
            Section::Bss(s) => Ok(s.data_size),
        }
    }
}

impl<'data, I> TryFrom<(&str, SectionInfo<I>)> for Section<I> {
    type Error = Error;

    fn try_from(value: (&str, SectionInfo<I>)) -> Result<Self, Self::Error> {
        if value.0.starts_with(".text") {
            Ok(Section::Text(value.1))
        } else if value.0.starts_with(".rodata") {
            Ok(Section::Data(value.1, true))
        } else if value.0.starts_with(".data") {
            Ok(Section::Data(value.1, false))
        } else if value.0.starts_with(".bss") {
            Ok(Section::Bss(value.1))
        } else {
            Err(Error::InvalidSectionName)
        }
    }
}

impl<I> Display for SectionInfo<I> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "file_size: {}, data_size: {}, data: {}, offset: {}",
            self.file_size,
            self.data_size,
            self.data_index.is_some(),
            self.offset,
        )
    }
}

impl<I> Display for Section<I> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            Section::Text(s) => writeln!(f, "TEXT {}", s),
            Section::Data(s, false) => writeln!(f, "DATA {}", s),
            Section::Data(s, true) => writeln!(f, "RO-DATA {}", s),
            Section::Bss(s) => writeln!(f, "BSS {}", s),
        }
    }
}
